from django import forms
from django import template

register = template.Library()


@register.filter(name='is_checkbox')
def is_checkbox(field):
    """
    Boolean filter for form fields to determine if a field is using a checkbox
    widget.
    """
    return isinstance(field.field.widget, forms.CheckboxInput)



import collections

from django import forms
from django.forms.utils import ErrorDict
from django.core.exceptions import NON_FIELD_ERRORS
from django.template.loader import render_to_string
import six

try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError:  # Python < 3.0, Django < 3.0
    def python_2_unicode_compatible(klass):
        return klass


class CSSClassMixin(object):
    """
    Sane defaults for error and css classes.
    """
    error_css_class = 'error'
    required_css_class = 'required'


class NonBraindamagedErrorMixin(object):
    """
    Form mixin for easier field based error messages.
    """
    def field_error(self, name, error):
        self._errors = self._errors or ErrorDict()
        self._errors.setdefault(name, self.error_class())
        self._errors[name].append(error)

    def form_error(self, error):
        self.field_error(NON_FIELD_ERRORS, error)


class LabelSuffixMixin(object):
    """
    Form mixin to make it possible to override the label_suffix at class
    declaration.  Django's built-in Form class only allows you to override the
    label_suffix at instantiation or by overriding __init__.  A value of None
    will use the default Django provided suffix (':' in English).  If you wish
    to have no label_suffix, you can set label_suffix to and empty string.  For
    example,

        class NoLabelSuffixMixin(LabelSuffixMixin):
            label_suffix = ''
    """
    label_suffix = None

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('label_suffix', self.label_suffix)
        super(LabelSuffixMixin, self).__init__(*args, **kwargs)


def process_fieldset_row(fields, fieldset_class, base_name):
    for index, row in enumerate(fields):
        if not isinstance(row, (six.string_types, Fieldset)):
            if len(row) == 2 and isinstance(row[0], six.string_types) and isinstance(row[1], dict):
                row = fieldset_class(row[0], **row[1])
            else:
                row = fieldset_class("{0}_{1}".format(base_name, index), fields=row)
        yield row


def flatten(elements):
    """
    Flattens a mixed list of strings and iterables of strings into a single
    iterable of strings.
    """
    for element in elements:
        if isinstance(element, collections.Iterable) and not isinstance(element, six.string_types):
            for sub_element in flatten(element):
                yield sub_element
        else:
            yield element

flatten_to_tuple = lambda x: tuple(flatten(x))


@python_2_unicode_compatible
class Fieldset(CSSClassMixin):
    FIELDSET_CSS_CLASS = 'formFieldset'
    css_classes = None
    template_name = None

    def __init__(self, name, fields=[], **kwargs):
        self.name = name
        self.base_fields = tuple(process_fieldset_row(fields, type(self), name))
        self.legend = kwargs.pop("legend", None)
        # Check for duplicate names.
        names = [str(thing) for thing in self.base_fields]
        duplicates = [x for x, y in collections.Counter(names).items() if y > 1]
        if duplicates:
            raise AttributeError('Name Conflict in fieldset `{0}`.  The name(s) `{1}` appear multiple times.'.format(self.name, duplicates))
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    def __iter__(self):
        return iter(self.base_fields)

    def __bool__(self):
        return bool(self.base_fields)

    # Python 2
    __nonzero__ = __bool__

    def __str__(self):
        return self.name

    @property
    def fields(self):
        return flatten_to_tuple(self)


class BoundFieldset(object):
    is_fieldset = True

    def __init__(self, form, fieldset, name):
        self.form = form
        self.name = name
        self.fieldset = fieldset
        self.rows = collections.OrderedDict()
        for row in fieldset:
            self.rows[six.text_type(row)] = row

    def __getitem__(self, key):
        """
        >>> fieldset[1]
        # returns the item at index-1 in the fieldset
        >>> fieldset['name']
        # returns the item in the fieldset under the key 'name'
        """
        if isinstance(key, int) and not key in self.rows:
            return self[list(self.rows.keys())[key]]
        value = self.rows[key]
        if isinstance(value, six.string_types):
            return self.form[value]
        else:
            return type(self)(self.form, value, key)

    def __str__(self):
        env = {
            'fieldset': self,
            'form': self.form,
            'fieldset_template_name': 'betterforms/fieldset_as_div.html',
        }
        # TODO: don't hardcode the default template name.
        return render_to_string(self.template_name or 'betterforms/fieldset_as_div.html', env)

    def __iter__(self):
        for name in self.rows.keys():
            yield self[name]

    @property
    def template_name(self):
        return self.fieldset.template_name

    @property
    def errors(self):
        return self.form.errors.get(self.name, self.form.error_class())

    @property
    def css_classes(self):
        css_classes = set((self.fieldset.FIELDSET_CSS_CLASS, self.name))
        css_classes.update(self.fieldset.css_classes or [])
        if self.errors:
            css_classes.add(self.fieldset.error_css_class)
        return ' '.join(css_classes)

    @property
    def legend(self):
        return self.fieldset.legend


class FieldsetMixin(NonBraindamagedErrorMixin):
    template_name = None
    fieldset_class = Fieldset
    bound_fieldset_class = BoundFieldset
    base_fieldsets = None

    @property
    def fieldsets(self):
        if self.base_fieldsets is None:
            return self.bound_fieldset_class(self, self.fields.keys(), '__base_fieldset__')
        return self.bound_fieldset_class(self, self.base_fieldsets, self.base_fieldsets.name)

    def __getitem__(self, key):
        try:
            return super(FieldsetMixin, self).__getitem__(key)
        except KeyError:
            return self.fieldsets[key]

    def __iter__(self):
        for fieldset in self.fieldsets:
            yield fieldset
        #return iter(self.fieldsets)

    # These methods need to be implemented to render the fieldsets and fields
    # in a similar structure as `BaseForm`
    def __str__(self):
        return self.as_table()

    def as_table(self):
        raise NotImplementedError('To be implemented')

    def as_ul(self):
        raise NotImplementedError('To be implemented')

    def as_p(self):
        env = {
            'form': self,
            'fieldset_template_name': 'betterforms/fieldset_as_p.html',
            'field_template_name': 'betterforms/field_as_p.html',
        }
        return render_to_string(self.template_name or 'betterforms/form_as_p.html', env)


def get_fieldsets(bases, attrs):
    try:
        return attrs['Meta'].fieldsets
    except (KeyError, AttributeError):
        for base in bases:
            fieldsets = getattr(base, 'base_fieldsets', None)
            if fieldsets is not None:
                return fieldsets
    return None


def get_fieldset_class(bases, attrs):
    if 'fieldset_class' in attrs:
        return attrs['fieldset_class']
    else:
        for base in bases:
            try:
                return base.fieldset_class
            except AttributeError:
                continue
        return Fieldset


class BetterModelFormMetaclass(forms.models.ModelFormMetaclass):
    def __new__(cls, name, bases, attrs):
        base_fieldsets = get_fieldsets(bases, attrs)
        if base_fieldsets is not None:
            FieldsetClass = get_fieldset_class(bases, attrs)
            base_fieldsets = FieldsetClass('__base_fieldset__', fields=base_fieldsets)
            attrs['base_fieldsets'] = base_fieldsets
            Meta = attrs.get('Meta')
            if Meta and Meta.__dict__.get('fields') is None and Meta.__dict__.get('exclude') is None:
                attrs['Meta'].fields = flatten_to_tuple(base_fieldsets)
        attrs['base_fieldsets'] = base_fieldsets
        return super(BetterModelFormMetaclass, cls).__new__(cls, name, bases, attrs)


class BetterModelForm(six.with_metaclass(BetterModelFormMetaclass, FieldsetMixin, LabelSuffixMixin, CSSClassMixin, forms.ModelForm)):
    pass


class BetterFormMetaClass(forms.forms.DeclarativeFieldsMetaclass):
    def __new__(cls, name, bases, attrs):
        base_fieldsets = get_fieldsets(bases, attrs)
        if base_fieldsets is not None:
            FieldsetClass = get_fieldset_class(bases, attrs)
            base_fieldsets = FieldsetClass('__base_fieldset__', fields=base_fieldsets)
        attrs['base_fieldsets'] = base_fieldsets
        return super(BetterFormMetaClass, cls).__new__(cls, name, bases, attrs)


class BetterForm(six.with_metaclass(BetterFormMetaClass, FieldsetMixin, LabelSuffixMixin, CSSClassMixin, forms.forms.BaseForm)):
    """
    A 'Better' base Form class.
    """



import copy

from django import forms
from django.forms.utils import pretty_name
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.db.models import Q
from collections import OrderedDict
import six
from six.moves import reduce
from django.utils.http import urlencode

from .forms import BetterForm


def construct_querystring(data, **kwargs):
    params = copy.copy(data)

    # We can't call update here because QueryDict extends rather than replaces.
    for key, value in kwargs.items():
        params[key] = value

    if hasattr(params, 'urlencode'):
        return params.urlencode()
    else:
        return urlencode(params)


class IterDict(OrderedDict):
    """
    Extension of djangos built in sorted dictionary class which iterates
    through the values rather than keys.
    """
    def __iter__(self):
        for key in super(IterDict, self).__iter__():
            yield self[key]


class BaseChangeListForm(BetterForm):
    """
    Base class for all ``ChangeListForms``.
    """
    def __init__(self, *args, **kwargs):
        """
        Takes an option named argument ``queryset`` as the base queryset used in
        the ``get_queryset`` method.
        """
        try:
            self.base_queryset = kwargs.pop('queryset', None)
            if self.base_queryset is None:
                self.base_queryset = self.model.objects.all()
        except AttributeError:
            raise AttributeError('`ChangeListForm`s must be instantiated with a\
                                 queryset, or have a `model` attribute set on\
                                 them')
        super(BaseChangeListForm, self).__init__(*args, **kwargs)

    def get_queryset(self):
        """
        If the form was initialized with a queryset, this method returns that
        queryset.  Otherwise it returns ``Model.objects.all()`` for whatever
        model was defined for the form.
        """
        return self.base_queryset


class SearchForm(BaseChangeListForm):
    SEARCH_FIELDS = None
    CASE_SENSITIVE = False
    q = forms.CharField(label="Search", required=False)

    def __init__(self, *args, **kwargs):
        self.SEARCH_FIELDS = kwargs.pop('search_fields', self.SEARCH_FIELDS)
        super(SearchForm, self).__init__(*args, **kwargs)

        if self.SEARCH_FIELDS is None:
            raise ImproperlyConfigured('`SearchForm`s must be instantiated with an\
                                       iterable of fields to search over, or have \
                                       a `SEARCH_FIELDS` attribute set on them.')

    def get_queryset(self):
        """
        Constructs an '__contains' or '__icontains' filter across all of the
        fields listed in ``SEARCH_FIELDS``.
        """
        qs = super(SearchForm, self).get_queryset()

        # Do Searching
        q = self.cleaned_data.get('q', '').strip()
        if q:
            args = []
            for field in self.SEARCH_FIELDS:
                if self.CASE_SENSITIVE:
                    kwarg = {field + '__contains': q}
                else:
                    kwarg = {field + '__icontains': q}
                args.append(Q(**kwarg))
            if len(args) > 1:
                qs = qs.filter(reduce(lambda x, y: x | y, args))
            elif len(args) == 1:
                qs = qs.filter(args[0])

        return qs


class BoundHeader(object):
    def __init__(self, form, header):
        self.form = form
        self.header = header
        self.sorts = getattr(form, 'cleaned_data', {}).get('sorts', [])
        self.param = "{0}-sorts".format(form.prefix or '').strip('-')

    @property
    def name(self):
        return self.header.name

    @property
    def label(self):
        return self.header.label

    @property
    def column_name(self):
        return self.header.column_name

    @property
    def is_sortable(self):
        return self.header.is_sortable

    @property
    def _index(self):
        return self.form.HEADERS.index(self.header)

    @property
    def _sort_index(self):
        """
        1-indexed value for what number represents this header in the sorts
        querystring parameter.
        """
        return self._index + 1

    @property
    def is_active(self):
        """
        Returns whether this header is currently being used for sorting.
        """
        return self._sort_index in map(abs, self.sorts)

    @property
    def is_ascending(self):
        """
        Returns whether this header is currently being used for sorting in
        ascending order.
        """
        return self.is_active and self._sort_index in self.sorts

    @property
    def is_descending(self):
        """
        Returns whether this header is currently being used for sorting in
        descending order.
        """
        return self.is_active and self._sort_index not in self.sorts

    @property
    def css_classes(self):
        """
        String suitable to be used for the `class` attribute for an HTML
        element.  Denotes whether this header is active in the sorts, and the
        order in which it is being used.
        """
        classes = []
        if self.is_active:
            classes.append('active')
            if self.is_ascending:
                classes.append('ascending')
            elif self.is_descending:
                classes.append('descending')
        return ' '.join(classes)

    def add_to_sorts(self):
        """
        Compute the sorts that should be used when we're clicked on. If we're
        currently in the sorts, we'll be set as the first sort [ascending].
        Unless we're already at the front then we'll be inverted.
        """
        if self.sorts and abs(self.sorts[0]) == self._sort_index:
            return [-1 * self.sorts[0]] + self.sorts[1:]
        else:
            return [self._sort_index] + list(filter(lambda x: abs(x) != self._sort_index, self.sorts))

    @property
    def priority(self):
        if self.is_active:
            return list(map(abs, self.sorts)).index(self._sort_index) + 1

    @property
    def querystring(self):
        return construct_querystring(self.form.data, **{self.param: '.'.join(map(str, self.add_to_sorts()))})

    @property
    def singular_querystring(self):
        if self.is_active and abs(self.sorts[0]) == self._sort_index:
            value = -1 * self._sort_index
        else:
            value = self._sort_index
        return construct_querystring(self.form.data, **{self.param: str(value)})

    @property
    def remove_querystring(self):
        return construct_querystring(self.form.data, **{self.param: '.'.join(map(str, self.add_to_sorts()[1:]))})


class Header(object):
    BoundClass = BoundHeader
    column_name = None

    def __init__(self, name, label=None, column_name=False, is_sortable=True):
        self.name = name
        self.label = label or pretty_name(name)
        if is_sortable:
            self.column_name = column_name or name
        self.is_sortable = is_sortable


def is_header_kwargs(header):
    try:
        if not len(header) == 2:
            return False
    except AttributeError:
        return False
    try:
        return all((
            isinstance(header[0], six.string_types),
            isinstance(header[1], dict),
        ))
    except (IndexError, KeyError):
        return False


class HeaderSet(object):
    HeaderClass = Header

    def __init__(self, form, headers):
        self.form = form
        self.headers = OrderedDict()
        if headers is None:
            return
        for header in headers:
            if isinstance(header, Header):
                self.headers[header.name] = header
            elif isinstance(header, six.string_types):
                self.headers[header] = self.HeaderClass(header)
            elif is_header_kwargs(header):
                header_name, header_kwargs = header
                self.headers[header_name] = self.HeaderClass(header_name, **header_kwargs)
            elif len(header):
                try:
                    header_name = header[0]
                    header_args = header[1:]
                    self.headers[header_name] = self.HeaderClass(header_name, *header_args)
                except KeyError:
                    raise ImproperlyConfigured('Unknown format in header declaration: `{0}`'.format(repr(header)))
            else:
                raise ImproperlyConfigured('Unknown format in header declaration: `{0}`'.format(repr(header)))
        if not len(self) == len(headers):
            raise ImproperlyConfigured('Header names must be unique')

    def __len__(self):
        return len(self.headers)

    def __iter__(self):
        for header in self.headers.values():
            yield self.HeaderClass.BoundClass(self.form, header)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.HeaderClass.BoundClass(self.form, list(self.headers.values())[key])
        else:
            return self.HeaderClass.BoundClass(self.form, self.headers[key])


class SortFormBase(BetterForm):
    """
    A base class for writing your own SortForm. This form handles everything
    except applying the sorts to the queryset, which is convenient if you
    aren't working within the ChangeListForm paradigm.

    Usage::

        class MyForm(SortFormBase):
            HEADERS = (
                Header('name', label='Name'),
            )

            # fields ...

            def get_results(self):
                queryset = # ...
                queryset = self.apply_sorting(queryset)
                return queryset
    """
    HeaderSetClass = HeaderSet
    error_messages = {
        'unknown_header': 'Invalid sort parameter',
        'unsortable_header': 'Invalid sort parameter',
    }
    HEADERS = None
    sorts = forms.CharField(required=False, widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs):
        super(SortFormBase, self).__init__(*args, **kwargs)
        self.headers = self.HeaderSetClass(self, self.HEADERS)

    def clean_sorts(self):
        cleaned_data = self.cleaned_data
        sorts = list(filter(bool, cleaned_data.get('sorts', '').split('.')))
        if not sorts:
            return []
        # Ensure that the sort parameter does not contain non-numeric sort indexes
        if not all([sort.strip('-').isdigit() for sort in sorts]):
            raise ValidationError(self.error_messages['unknown_header'])
        sorts = [int(sort) for sort in sorts]
        # Ensure that all of our sort parameters are in range of our header values
        if any([abs(sort) > len(self.HEADERS) for sort in sorts]):
            raise ValidationError(self.error_messages['unknown_header'])
        # Ensure not un-sortable fields are being sorted by
        if not all(self.HEADERS[abs(i) - 1].is_sortable for i in sorts):
            raise ValidationError(self.error_messages['unsortable_header'])

        return sorts

    def get_order_by(self):
        # Do Sorting
        sorts = self.cleaned_data.get('sorts', [])
        order_by = []
        for sort in sorts:
            param = self.headers[abs(sort) - 1].column_name
            if sort < 0:
                param = '-' + param
            order_by.append(param)
        return order_by

    def apply_sorting(self, qs):
        order_by = self.get_order_by()
        if order_by:
            qs = qs.order_by(*order_by)
        return qs


class SortForm(BaseChangeListForm, SortFormBase):
    def get_queryset(self):
        """
        Returns an ordered queryset, sorted based on the values submitted in
        the sort parameter.
        """
        qs = super(SortForm, self).get_queryset()
        return self.apply_sorting(qs)



import sys
if sys.version_info < (2, 7):
    # In python < 2.7 unittest doesn't have expectedFailure
    from django.utils import unittest
else:
    # django.utils.unittest became deprecated in Django 1.7
    import unittest  # NOQA

import mock

import django
from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.db import models
from django.test import TestCase
from django.template.loader import render_to_string
from django.http import QueryDict

from betterforms.changelist import (
    BaseChangeListForm, SearchForm, SortForm, HeaderSet, Header, BoundHeader
)
from betterforms.forms import (
    BetterForm, BetterModelForm, Fieldset, BoundFieldset, flatten_to_tuple,
)


class TestUtils(TestCase):
    def test_flatten(self):
        fields1 = ('a', 'b', 'c')
        self.assertTupleEqual(flatten_to_tuple(fields1), fields1)

        fields2 = ('a', ('b', 'c'), 'd')
        self.assertTupleEqual(flatten_to_tuple(fields2), ('a', 'b', 'c', 'd'))

        fields3 = ('a', ('b', 'c'), 'd', ('e', ('f', 'g', ('h',)), 'i'))
        self.assertTupleEqual(flatten_to_tuple(fields3), ('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i'))


class TestFieldSets(TestCase):
    def test_basic_fieldset(self):
        fields = ('a', 'b', 'c')
        fieldset = Fieldset('the_name', fields=fields)
        self.assertEqual(fieldset.name, 'the_name')
        self.assertTupleEqual(fields, fieldset.fields)

    def test_nested_fieldset(self):
        fields = ('a', ('b', 'c'), 'd')
        fieldset = Fieldset('the_name', fields=fields)
        self.assertTupleEqual(flatten_to_tuple(fields), fieldset.fields)
        iterated = tuple(iter(fieldset))
        self.assertEqual(iterated[0], 'a')
        self.assertTupleEqual(iterated[1].fields, ('b', 'c'))
        self.assertEqual(iterated[2], 'd')

    def test_named_nested_fieldset(self):
        fields = ('a', ('sub_name', {'fields': ('b', 'c')}), 'd')
        fieldset = Fieldset('the_name', fields=fields)
        self.assertTupleEqual(fieldset.fields, ('a', 'b', 'c', 'd'))
        fieldsets = tuple(iter(fieldset))
        self.assertEqual(fieldsets[0], 'a')
        self.assertTupleEqual(fieldsets[1].fields, ('b', 'c'))
        self.assertEqual(fieldsets[1].name, 'sub_name')
        self.assertEqual(fieldsets[2], 'd')

    def test_deeply_nested_fieldsets(self):
        fields = ('a', ('b', 'c'), 'd', ('e', ('f', 'g', ('h',)), 'i'))
        fieldset = Fieldset('the_name', fields=fields)
        self.assertTupleEqual(flatten_to_tuple(fields), fieldset.fields)

    def test_fieldset_as_row_item(self):
        fields = ('a', Fieldset('sub_name', fields=['b', 'c']))
        fieldset = Fieldset('the_name', fields=fields)
        self.assertTupleEqual(fieldset.fields, ('a', 'b', 'c'))

    def test_nonzero_fieldset(self):
        fieldset1 = Fieldset('the_name', fields=[])
        self.assertFalse(fieldset1)

        fieldset2 = Fieldset('the_name', fields=['a'])
        self.assertTrue(fieldset2)

    def test_assigning_template_name(self):
        fieldset1 = Fieldset('the_name', fields=['a'])
        self.assertIsNone(fieldset1.template_name)
        fieldset2 = Fieldset('the_name', fields=['a'], template_name='some_custom_template.html')
        self.assertEqual(fieldset2.template_name, 'some_custom_template.html')


class TestFieldsetDeclarationSyntax(TestCase):
    def test_admin_style_declaration(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()
            d = forms.CharField()

            class Meta:
                fieldsets = (
                    ('first', {'fields': ('a')}),
                    ('second', {'fields': ('b', 'c')}),
                    ('third', {'fields': ('d')}),
                )
        form = TestForm()
        fieldsets = [fieldset for fieldset in form.fieldsets]
        self.assertEqual(fieldsets[0].name, 'first')
        self.assertTupleEqual(fieldsets[0].fieldset.fields, ('a',))
        self.assertEqual(fieldsets[1].name, 'second')
        self.assertTupleEqual(fieldsets[1].fieldset.fields, ('b', 'c'))
        self.assertEqual(fieldsets[2].name, 'third')
        self.assertTupleEqual(fieldsets[2].fieldset.fields, ('d',))
        self.assertIsInstance(fieldsets[0], BoundFieldset)

    def test_bare_fields_style_declaration(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()
            d = forms.CharField()

            class Meta:
                fieldsets = ('a', ('b', 'c'), 'd')
        form = TestForm()
        fieldsets = [fieldset for fieldset in form.fieldsets]
        self.assertEqual(fieldsets[0].field, form.fields['a'])
        self.assertEqual(fieldsets[1].name, '__base_fieldset___1')
        self.assertTupleEqual(fieldsets[1].fieldset.fields, ('b', 'c'))
        self.assertEqual(fieldsets[2].field, form.fields['d'])
        self.assertIsInstance(fieldsets[0], forms.forms.BoundField)
        self.assertIsInstance(fieldsets[1], BoundFieldset)
        self.assertIsInstance(fieldsets[2], forms.forms.BoundField)


class TestBetterForm(TestCase):
    def setUp(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()

            class Meta:
                fieldsets = (
                    ('first', {'fields': ('a', 'b')}),
                    ('second', {'fields': ('c')}),
                )
        self.TestForm = TestForm

    def test_name_lookups(self):
        form = self.TestForm()
        fieldsets = [fieldset for fieldset in form.fieldsets]
        # field lookups
        self.assertEqual(form['a'].field, fieldsets[0]['a'].field)
        # fieldset lookups
        self.assertEqual(form['first'].fieldset, fieldsets[0].fieldset)
        self.assertEqual(form['second'].fieldset, fieldsets[1].fieldset)

    def test_index_lookups(self):
        form = self.TestForm()
        # field lookups
        self.assertEqual(form['a'].field, form.fieldsets[0][0].field)
        # fieldset lookups
        self.assertEqual(form['first'].fieldset, form.fieldsets[0].fieldset)
        self.assertEqual(form['second'].fieldset, form.fieldsets[1].fieldset)

    def test_field_to_fieldset_name_conflict(self):
        with self.assertRaises(AttributeError):
            class NameConflictForm(self.TestForm):
                class Meta:
                    fieldsets = (
                        ('first', {'fields': ('a', 'b')}),
                        ('first', {'fields': ('c')}),
                    )

    def test_duplicate_name_in_fieldset(self):
        with self.assertRaises(AttributeError):
            class NameConflictForm(self.TestForm):
                class Meta:
                    fieldsets = (
                        ('first', {'fields': ('a', 'a')}),
                        ('second', {'fields': ('c')}),
                    )

    def test_field_error(self):
        data = {'a': 'a', 'b': 'b', 'c': 'c'}
        form = self.TestForm(data)
        self.assertTrue(form.is_valid())

        form.field_error('a', 'test')
        self.assertFalse(form.is_valid())

    def test_form_error(self):
        data = {'a': 'a', 'b': 'b', 'c': 'c'}
        form = self.TestForm(data)
        self.assertTrue(form.is_valid())

        form.form_error('test')
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors, {'__all__': [u'test']})

    def test_fieldset_error(self):
        data = {'a': 'a', 'b': 'b', 'c': 'c'}
        form = self.TestForm(data)
        self.assertTrue(form.is_valid())

        self.assertNotIn(form.fieldsets[0].fieldset.error_css_class, form.fieldsets[0].css_classes)

        form.field_error('first', 'test')
        self.assertFalse(form.is_valid())
        fieldsets = [fieldset for fieldset in form.fieldsets]
        self.assertTrue(fieldsets[0].errors)
        self.assertIn(form.fieldsets[0].fieldset.error_css_class, form.fieldsets[0].css_classes)

    def test_fieldset_css_classes(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()

            class Meta:
                fieldsets = (
                    ('first', {'fields': ('a', 'b')}),
                    ('second', {'fields': ('c'), 'css_classes': ['arst', 'tsra']}),
                )
        form = TestForm()
        self.assertIn('arst', form.fieldsets[1].css_classes)
        self.assertIn('tsra', form.fieldsets[1].css_classes)

    def test_fieldset_iteration(self):
        form = self.TestForm()
        self.assertTupleEqual(
            tuple(fieldset.fieldset for fieldset in form),
            tuple(fieldset.fieldset for fieldset in form.fieldsets),
        )

    def test_no_fieldsets(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()

        form = TestForm()
        fields_iter = sorted((field.field for field in form), key=id)
        fields_values = sorted(form.fields.values(), key=id)
        self.assertSequenceEqual(fields_iter, fields_values)


class TestBetterModelForm(TestCase):
    def setUp(self):
        class TestModel(models.Model):
            class Meta:
                abstract = True
            a = models.CharField(max_length=255)
            b = models.CharField(max_length=255)
            c = models.CharField(max_length=255)
            d = models.CharField(max_length=255)
        self.TestModel = TestModel

    def test_basic_fieldsets(self):
        class TestModelForm(BetterModelForm):
            class Meta:
                model = self.TestModel
                fieldsets = (
                    ('first', {'fields': ('a',)}),
                    ('second', {'fields': ('b', 'c')}),
                    ('third', {'fields': ('d',)}),
                )
        form = TestModelForm()
        fieldsets = [fieldset for fieldset in form.fieldsets]
        self.assertEqual(fieldsets[0].name, 'first')
        self.assertEqual(fieldsets[1].name, 'second')
        self.assertEqual(fieldsets[2].name, 'third')
        self.assertTupleEqual(fieldsets[0].fieldset.fields, ('a',))
        self.assertTupleEqual(fieldsets[1].fieldset.fields, ('b', 'c'))
        self.assertTupleEqual(fieldsets[2].fieldset.fields, ('d',))

    def test_fields_meta_attribute(self):
        class TestModelForm1(BetterModelForm):
            class Meta:
                model = self.TestModel
                fieldsets = (
                    ('first', {'fields': ('a',)}),
                    ('second', {'fields': ('b', 'c')}),
                    ('third', {'fields': ('d',)}),
                )
        self.assertTrue(hasattr(TestModelForm1.Meta, 'fields'))
        self.assertTupleEqual(TestModelForm1.Meta.fields, ('a', 'b', 'c', 'd'))

        class TestModelForm2(BetterModelForm):
            class Meta:
                model = self.TestModel
                fieldsets = (
                    ('first', {'fields': ('a',)}),
                    ('second', {'fields': ('b', 'c')}),
                    ('third', {'fields': ('d',)}),
                )
                fields = ('a', 'b', 'd')

        self.assertTrue(hasattr(TestModelForm2.Meta, 'fields'))
        self.assertTupleEqual(TestModelForm2.Meta.fields, ('a', 'b', 'd'))

        class TestModelForm3(TestModelForm2):
            pass

        self.assertTrue(hasattr(TestModelForm3.Meta, 'fields'))
        self.assertTupleEqual(TestModelForm3.Meta.fields, ('a', 'b', 'd'))

        class TestModelForm4(TestModelForm2):
            class Meta(TestModelForm2.Meta):
                fieldsets = (
                    ('first', {'fields': ('a', 'c')}),
                    ('third', {'fields': ('d',)}),
                )

        self.assertTrue(hasattr(TestModelForm4.Meta, 'fields'))
        self.assertTupleEqual(TestModelForm4.Meta.fields, ('a', 'c', 'd'))


class TestFormRendering(TestCase):
    def setUp(self):
        class TestForm(BetterForm):
            # Set the label_suffix to an empty string for consistent results
            # across Django 1.5 and 1.6.
            label_suffix = ''

            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()

            class Meta:
                fieldsets = (
                    ('first', {'fields': ('a', 'b')}),
                    ('second', {'fields': ('c')}),
                )
        self.TestForm = TestForm

    def test_non_fieldset_form_rendering(self):
        class TestForm(BetterForm):
            # Set the label_suffix to an empty string for consistent results
            # across Django 1.5 and 1.6.
            label_suffix = ''

            a = forms.CharField()
            b = forms.CharField(required=False)
            c = forms.CharField()

        form = TestForm()
        env = {
            'form': form,
            'no_head': True,
            'fieldset_template_name': 'betterforms/fieldset_as_div.html',
            'field_template_name': 'betterforms/field_as_div.html',
        }
        test = """
            <div class="required a formField">
                <label class="required" for="id_a">A</label>
                <input id="id_a" name="a" required type="text" />
            </div>
            <div class="b formField">
                <label for="id_b">B</label>
                <input id="id_b" name="b" type="text" />
            </div>
            <div class="required c formField">
                <label class="required" for="id_c">C</label>
                <input id="id_c" name="c" required type="text" />
            </div>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )
        form.field_error('a', 'this is an error message')
        test = """
            <div class="required error a formField">
                <label class="required" for="id_a">A</label>
                <input id="id_a" name="a" required type="text" />
                <ul class="errorlist"><li>this is an error message</li></ul>
            </div>
            <div class="b formField">
                <label for="id_b">B</label>
                <input id="id_b" name="b" type="text" />
            </div>
            <div class="required c formField">
                <label class="required" for="id_c">C</label>
                <input id="id_c" name="c" required type="text" />
            </div>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )

    def test_include_tag_rendering(self):
        form = self.TestForm()
        env = {
            'form': form,
            'no_head': True,
            'fieldset_template_name': 'betterforms/fieldset_as_div.html',
            'field_template_name': 'betterforms/field_as_div.html',
        }
        test = """
            <fieldset class="formFieldset first">
                <div class="required a formField">
                    <label class="required" for="id_a">A</label>
                    <input id="id_a" name="a" required type="text" />
                </div>
                <div class="required b formField">
                    <label class="required" for="id_b">B</label>
                    <input id="id_b" name="b" required type="text" />
                </div>
            </fieldset>
            <fieldset class="formFieldset second">
                <div class="required c formField">
                    <label class="required" for="id_c">C</label>
                    <input id="id_c" name="c" required type="text" />
                </div>
            </fieldset>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )
        form.field_error('a', 'this is an error message')
        test = """
            <fieldset class="formFieldset first">
                <div class="required error a formField">
                    <label class="required" for="id_a">A</label>
                    <input id="id_a" name="a" required type="text" />
                    <ul class="errorlist"><li>this is an error message</li></ul>
                </div>
                <div class="required b formField">
                    <label class="required" for="id_b">B</label>
                    <input id="id_b" name="b" required type="text" />
                </div>
            </fieldset>
            <fieldset class="formFieldset second">
                <div class="required c formField">
                    <label class="required" for="id_c">C</label>
                    <input id="id_c" name="c" required type="text" />
                </div>
            </fieldset>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )

    def test_fields_django_form_required(self):
        class TestForm(forms.Form):
            a = forms.CharField(label='A:')
            b = forms.CharField(label='B:', required=False)

        form = TestForm()

        env = {
            'form': form,
            'no_head': True,
            'fieldset_template_name': 'betterforms/fieldset_as_div.html',
            'field_template_name': 'betterforms/field_as_div.html',
        }
        test = """
            <div class="a formField required">
                <label for="id_a">A:</label>
                <input id="id_a" name="a" required type="text" />
            </div>
            <div class="b formField">
                <label for="id_b">B:</label>
                <input id="id_b" name="b" type="text" />
            </div>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )

    @unittest.expectedFailure
    def test_form_to_str(self):
        # TODO: how do we test this
        assert False

    @unittest.expectedFailure
    def test_form_as_table(self):
        form = self.TestForm()
        form.as_table()

    @unittest.expectedFailure
    def test_form_as_ul(self):
        form = self.TestForm()
        form.as_ul()

    def test_form_as_p(self):
        form = self.TestForm()
        test = """
            <fieldset class="formFieldset first">
                <p class="required">
                    <label class="required" for="id_a">A</label>
                    <input id="id_a" name="a" required type="text" />
                </p>
                <p class="required">
                    <label class="required" for="id_b">B</label>
                    <input id="id_b" name="b" required type="text" />
                </p>
            </fieldset>
            <fieldset class="formFieldset second">
                <p class="required">
                    <label class="required" for="id_c">C</label>
                    <input id="id_c" name="c" required type="text" />
                </p>
            </fieldset>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            form.as_p(),
            test,
        )

        form.field_error('a', 'this is an error')
        test = """
            <fieldset class="formFieldset first">
                <p class="required error">
                    <ul class="errorlist"><li>this is an error</li></ul>
                    <label class="required" for="id_a">A</label>
                    <input id="id_a" name="a" required type="text" />
                </p>
                <p class="required">
                    <label class="required" for="id_b">B</label>
                    <input id="id_b" name="b" required type="text" />
                </p>
            </fieldset>
            <fieldset class="formFieldset second">
                <p class="required">
                    <label class="required" for="id_c">C</label>
                    <input id="id_c" name="c" required type="text" />
                </p>
            </fieldset>
            """
        self.maxDiff=None
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            form.as_p(),
            test,
        )

    def test_fieldset_legend(self):
        class TestForm(BetterForm):
            a = forms.CharField()
            b = forms.CharField()
            c = forms.CharField()

            label_suffix = ''

            class Meta:
                fieldsets = (
                    Fieldset('first', ('a', 'b'), legend='First Fieldset'),
                    Fieldset('second', ('c'), legend='Second Fieldset'),
                )

        form = TestForm()
        test = """
            <fieldset class="formFieldset first">
                <legend>First Fieldset</legend>
                <p class="required">
                    <label class="required" for="id_a">A</label>
                    <input id="id_a" name="a" required type="text" />
                </p>
                <p class="required">
                    <label class="required" for="id_b">B</label>
                    <input id="id_b" name="b" required type="text" />
                </p>
            </fieldset>
            <fieldset class="formFieldset second">
                <legend>Second Fieldset</legend>
                <p class="required">
                    <label class="required" for="id_c">C</label>
                    <input id="id_c" name="c" required type="text" />
                </p>
            </fieldset>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            form.as_p(),
            test,
        )

    def test_css_classes_when_form_has_prefix(self):
        class TestForm(BetterForm):
            name = forms.CharField()
            label_suffix = ''

        form = TestForm(prefix="prefix")
        env = {'form': form, 'no_head': True}
        test = """
            <div class="required prefix-name name formField">
                <label class="required" for="id_prefix-name">Name</label>
                <input type="text" id="id_prefix-name" required name="prefix-name" />
            </div>
            """
        if django.VERSION < (1, 8):
            test = test.replace('label class="required"', 'label')
        if django.VERSION < (1, 10):
            test = test.replace(' required ', ' ')
        self.assertHTMLEqual(
            render_to_string('betterforms/form_as_fieldsets.html', env),
            test,
        )


class ChangeListModel(models.Model):
    field_a = models.CharField(max_length=255)
    field_b = models.CharField(max_length=255)
    field_c = models.TextField(max_length=255)


class TestChangleListQuerySetAPI(TestCase):
    def setUp(self):
        class TestChangeListForm(BaseChangeListForm):
            model = ChangeListModel
            foo = forms.CharField()
        self.TestChangeListForm = TestChangeListForm

        for i in range(5):
            ChangeListModel.objects.create(field_a=str(i))

    def test_with_model_declared(self):
        form = self.TestChangeListForm({})

        # base_queryset should default to Model.objects.all()
        self.assertTrue(form.base_queryset.count(), 5)

    def test_with_model_declaration_and_provided_queryset(self):
        form = self.TestChangeListForm({'foo': 'arst'}, queryset=ChangeListModel.objects.exclude(field_a='0').exclude(field_a='1'))

        self.assertTrue(form.base_queryset.count(), 3)
        form.full_clean()
        self.assertTrue(form.get_queryset().count(), 3)

    def test_missing_model_and_queryset(self):
        class TestChangeListForm(BaseChangeListForm):
            pass

        with self.assertRaises(AttributeError):
            TestChangeListForm()


class TestSearchFormAPI(TestCase):
    def setUp(self):
        ChangeListModel.objects.create(field_a='foo', field_b='bar', field_c='baz')
        ChangeListModel.objects.create(field_a='bar', field_b='baz')
        ChangeListModel.objects.create(field_a='baz')

    def test_requires_search_fields(self):
        class TheSearchForm(SearchForm):
            model = ChangeListModel

        with self.assertRaises(ImproperlyConfigured):
            TheSearchForm({})

    def test_passing_in_search_fields(self):
        class TheSearchForm(SearchForm):
            model = ChangeListModel

        form = TheSearchForm({}, search_fields=('field_a',))
        self.assertEqual(form.SEARCH_FIELDS, ('field_a',))

        form = TheSearchForm({}, search_fields=('field_a', 'field_b'))
        self.assertEqual(form.SEARCH_FIELDS, ('field_a', 'field_b'))

    def test_setting_search_fields_on_class(self):
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a', 'field_b', 'field_c')
            model = ChangeListModel

        form = TheSearchForm({})
        self.assertEqual(form.SEARCH_FIELDS, ('field_a', 'field_b', 'field_c'))

    def test_overriding_search_fields_set_on_class(self):
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a', 'field_b', 'field_c')
            model = ChangeListModel

        form = TheSearchForm({}, search_fields=('field_a', 'field_c'))
        self.assertEqual(form.SEARCH_FIELDS, ('field_a', 'field_c'))

    def test_searching(self):
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a', 'field_b', 'field_c')
            model = ChangeListModel

        f = TheSearchForm({'q': 'foo'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 1)

        f = TheSearchForm({'q': 'bar'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 2)

        f = TheSearchForm({'q': 'baz'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 3)

    def test_searching_over_limited_fields(self):
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a', 'field_c')
            model = ChangeListModel

        f = TheSearchForm({'q': 'foo'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 1)

        f = TheSearchForm({'q': 'bar'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 1)

        f = TheSearchForm({'q': 'baz'})
        f.full_clean()
        self.assertEqual(f.get_queryset().count(), 2)

    def test_case_insensitive(self):
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a',)
            model = ChangeListModel

        self.assertFalse(TheSearchForm.CASE_SENSITIVE)

        upper_cased = ChangeListModel.objects.create(field_a='TeSt')
        lower_cased = ChangeListModel.objects.create(field_a='test')

        form = TheSearchForm({'q': 'Test'})
        form.full_clean()

        self.assertIn(upper_cased, form.get_queryset())
        self.assertIn(lower_cased, form.get_queryset())

    @unittest.skipIf(settings.DATABASES['default']['ENGINE'] == 'django.db.backends.sqlite3', 'Case Sensitive __contains queries are not supported on sqlite.')
    def test_case_sensitive(self):
        # TODO: make this test run on travis with postgres/mysql to be sure
        # this functionality actually works.
        class TheSearchForm(SearchForm):
            SEARCH_FIELDS = ('field_a', 'field_c')
            CASE_SENSITIVE = True
            model = ChangeListModel

        self.assertTrue(TheSearchForm.CASE_SENSITIVE)

        upper_cased = ChangeListModel.objects.create(field_a='TeSt')
        lower_cased = ChangeListModel.objects.create(field_a='test')

        form = TheSearchForm({'q': 'TeSt'})
        form.full_clean()

        self.assertIn(upper_cased, form.get_queryset())
        self.assertNotIn(lower_cased, form.get_queryset())


class TestHeaderAPI(TestCase):
    def test_header_bare_declaration(self):
        header = Header('field_a')

        self.assertTrue(header.is_sortable)
        self.assertEqual(header.name, 'field_a')
        self.assertEqual(header.column_name, 'field_a')
        self.assertEqual(header.label, 'Field a')

    def test_header_with_label_declaration(self):
        header = Header('field_a', 'Test Label')

        self.assertEqual(header.label, 'Test Label')

    def test_header_with_column_declared(self):
        header = Header('not_a_column', column_name='field_a')

        self.assertEqual(header.name, 'not_a_column')
        self.assertEqual(header.column_name, 'field_a')

    def test_non_sortable_header(self):
        header = Header('field_a', is_sortable=False)

        self.assertFalse(header.is_sortable)


class TestHeaderSetAPI(TestCase):
    def test_header_names_must_be_unique(self):
        HEADERS = (
            Header('field_a'),
            Header('field_a'),
        )
        with self.assertRaises(ImproperlyConfigured):
            HeaderSet(None, HEADERS)

    def test_header_set_declared_as_header_classes(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b', 'Test Label'),
            Header('test_name', 'Test Name', column_name='field_c'),
            Header('created_at', is_sortable=False),
        )
        self.do_header_set_assertions(HEADERS)

    def test_header_set_declared_as_args(self):
        HEADERS = (
            ('field_a',),
            ('field_b', 'Test Label'),
            ('test_name', 'Test Name', 'field_c'),
            ('created_at', None, None, False),
        )
        self.do_header_set_assertions(HEADERS)

    def test_header_set_declared_as_name_and_kwargs(self):
        HEADERS = (
            ('field_a', {}),
            ('field_b', {'label': 'Test Label'}),
            ('test_name', {'label': 'Test Name', 'column_name': 'field_c'}),
            ('created_at', {'is_sortable': False}),
        )
        self.do_header_set_assertions(HEADERS)

    def test_header_set_mixed_declaration_styles(self):
        HEADERS = (
            'field_a',
            ('field_b', 'Test Label'),
            ('test_name', {'label': 'Test Name', 'column_name': 'field_c'}),
            Header('created_at', is_sortable=False),
        )
        self.do_header_set_assertions(HEADERS)

    def test_bad_header_declaration(self):
        HEADERS = (
            {'bad_declaration': 'test'},
            ('field_b', 'Test Label'),
            ('test_name', {'label': 'Test Name', 'column_name': 'field_c'}),
            Header('created_at', is_sortable=False),
        )
        with self.assertRaises(ImproperlyConfigured):
            self.do_header_set_assertions(HEADERS)

    def do_header_set_assertions(self, HEADERS):
        header_set = HeaderSet(None, HEADERS)

        self.assertTrue(len(header_set), 4)
        self.assertSequenceEqual(
            [header.name for header in header_set.headers.values()],
            ('field_a', 'field_b', 'test_name', 'created_at'),
        )
        self.assertSequenceEqual(
            [header.label for header in header_set.headers.values()],
            ('Field a', 'Test Label', 'Test Name', 'Created at'),
        )
        self.assertSequenceEqual(
            [header.column_name for header in header_set.headers.values()],
            ('field_a', 'field_b', 'field_c', None),
        )
        self.assertSequenceEqual(
            [header.is_sortable for header in header_set.headers.values()],
            (True, True, True, False),
        )

    def test_iteration_yields_bound_headers(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b', 'Test Label'),
            Header('test_name', 'Test Name', column_name='field_c'),
            Header('created_at', is_sortable=False),
        )
        form = mock.NonCallableMagicMock(forms.Form)
        form.prefix = None
        header_set = HeaderSet(form, HEADERS)

        self.assertTrue(all((
            isinstance(header, BoundHeader) for header in header_set
        )))

    def test_index_and_key_lookups(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b', 'Test Label'),
            Header('test_name', 'Test Name', column_name='field_c'),
            Header('created_at', is_sortable=False),
        )
        form = mock.NonCallableMagicMock(forms.Form)
        form.prefix = None
        header_set = HeaderSet(form, HEADERS)

        self.assertIsInstance(header_set[0], BoundHeader)
        self.assertEqual(header_set[0].header, HEADERS[0])

        self.assertIsInstance(header_set['field_b'], BoundHeader)
        self.assertEqual(header_set['field_b'].header, HEADERS[1])


class TestBoundHeaderAPI(TestCase):
    def setUp(self):
        self.HEADERS = (
            Header('test_name', 'Test Label', 'column_name', is_sortable=True),
        )
        self.form = mock.NonCallableMagicMock(forms.Form)
        self.form.prefix = None
        self.form.HEADERS = self.HEADERS

    def test_bound_header_pass_through_properties(self):
        header_set = HeaderSet(self.form, self.HEADERS)

        self.assertEqual(header_set[0].header, self.HEADERS[0])
        self.assertEqual(header_set[0].name, self.HEADERS[0].name)
        self.assertEqual(header_set[0].label, self.HEADERS[0].label)
        self.assertEqual(header_set[0].column_name, self.HEADERS[0].column_name)
        self.assertEqual(header_set[0].is_sortable, self.HEADERS[0].is_sortable)

    def test_sort_parameter_no_prefix(self):
        self.assertIsNone(self.form.prefix)
        header = HeaderSet(self.form, self.HEADERS)[0]

        self.assertEqual(header.param, 'sorts')

    def test_sort_parameter_with_prefix(self):
        self.form.prefix = 'test'
        header = HeaderSet(self.form, self.HEADERS)[0]

        self.assertEqual(header.param, 'test-sorts')

    def test_is_active_property_while_not_active(self):
        header = HeaderSet(self.form, self.HEADERS)[0]

        self.assertFalse(header.is_active)
        self.assertFalse(header.is_ascending)
        self.assertFalse(header.is_descending)

    def test_is_active_property_while_active_and_ascending(self):
        self.form.data = {'sorts': '1'}
        self.form.cleaned_data = {'sorts': [1]}
        header = HeaderSet(self.form, self.HEADERS)[0]

        self.assertTrue(header.is_active)
        self.assertTrue(header.is_ascending)
        self.assertFalse(header.is_descending)

    def test_is_active_property_while_active_and_descending(self):
        self.form.data = {'sorts': '-1'}
        self.form.cleaned_data = {'sorts': [-1]}
        header = HeaderSet(self.form, self.HEADERS)[0]

        self.assertTrue(header.is_active)
        self.assertFalse(header.is_ascending)
        self.assertTrue(header.is_descending)

    def test_add_to_sorts_with_no_sorts(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b'),
            Header('field_c'),
        )
        self.form.data = {}
        self.form.cleaned_data = {'sorts': []}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)
        self.assertEqual(header_set['field_a'].add_to_sorts(), [1])
        self.assertEqual(header_set['field_b'].add_to_sorts(), [2])
        self.assertEqual(header_set['field_c'].add_to_sorts(), [3])

    def test_add_to_sorts_with_active_sorts(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b'),
            Header('field_c'),
        )
        self.form.data = {'sorts': '1.-2'}
        self.form.cleaned_data = {'sorts': [1, -2]}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)
        self.assertEqual(header_set['field_a'].add_to_sorts(), [-1, -2])
        self.assertEqual(header_set['field_b'].add_to_sorts(), [2, 1])
        self.assertEqual(header_set['field_c'].add_to_sorts(), [3, 1, -2])

    def test_sort_priority_display(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b'),
            Header('field_c'),
        )
        self.form.data = {'sorts': '-2.1'}
        self.form.cleaned_data = {'sorts': [-2, 1]}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)
        self.assertEqual(header_set['field_a'].priority, 2)
        self.assertEqual(header_set['field_b'].priority, 1)
        self.assertEqual(header_set['field_c'].priority, None)

    def test_css_classes(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b'),
            Header('field_c'),
        )
        self.form.data = {'sorts': '1.-2'}
        self.form.cleaned_data = {'sorts': [1, -2]}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)
        self.assertEqual(header_set['field_a'].css_classes, 'active ascending')
        self.assertEqual(header_set['field_b'].css_classes, 'active descending')
        self.assertEqual(header_set['field_c'].css_classes, '')

    def test_bound_header_querystring_properties(self):
        HEADERS = (
            Header('field_a'),
            Header('field_b'),
            Header('field_c'),
        )
        self.form.data = {'sorts': '1.-2'}
        self.form.cleaned_data = {'sorts': [1, -2]}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)

        self.assertEqual(header_set['field_a'].querystring, 'sorts=-1.-2')
        self.assertEqual(header_set['field_a'].singular_querystring, 'sorts=-1')
        self.assertEqual(header_set['field_a'].remove_querystring, 'sorts=-2')

        self.assertEqual(header_set['field_b'].querystring, 'sorts=2.1')
        self.assertEqual(header_set['field_b'].singular_querystring, 'sorts=2')
        self.assertEqual(header_set['field_b'].remove_querystring, 'sorts=1')

        self.assertEqual(header_set['field_c'].querystring, 'sorts=3.1.-2')
        self.assertEqual(header_set['field_c'].singular_querystring, 'sorts=3')
        self.assertEqual(header_set['field_c'].remove_querystring, 'sorts=1.-2')

    def assertQueryStringEqual(self, a, b, *args, **kwargs):
        """
        We need this because QueryDicts are dicts and key-ordering is not
        guaranteed on Python3. So we just convert query_strings back into
        QueryDicts to compare them.
        """
        self.assertEqual(QueryDict(a), QueryDict(b), *args, **kwargs)

    def test_bound_header_querystring_with_querydict(self):
        HEADERS = (
            Header('field_a'),
        )
        self.form.data = QueryDict('field=value')
        self.form.cleaned_data = {'field': 'value'}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)

        self.assertQueryStringEqual(header_set['field_a'].querystring, 'field=value&sorts=1')
        self.assertQueryStringEqual(header_set['field_a'].singular_querystring, 'field=value&sorts=1')
        self.assertQueryStringEqual(header_set['field_a'].remove_querystring, 'field=value&sorts=')

    def test_bound_header_querystring_with_querydict_overwrites_instead_of_appending(self):
        HEADERS = (
            Header('field_a'),
        )
        self.form.data = QueryDict('field=value&sorts=1')
        self.form.cleaned_data = {'field': 'value', 'sorts': [1]}
        self.form.HEADERS = HEADERS
        header_set = HeaderSet(self.form, HEADERS)

        # It used to output 'field=value&sorts=1&sorts=-1'
        self.assertQueryStringEqual(header_set['field_a'].querystring, 'field=value&sorts=-1')
        self.assertQueryStringEqual(header_set['field_a'].singular_querystring, 'field=value&sorts=-1')
        self.assertQueryStringEqual(header_set['field_a'].remove_querystring, 'field=value&sorts=')


class TestSortFormAPI(TestCase):
    def setUp(self):
        self.abc = ChangeListModel.objects.create(field_a='a', field_b='b', field_c='c')
        self.cab = ChangeListModel.objects.create(field_a='c', field_b='a', field_c='b')
        self.bca = ChangeListModel.objects.create(field_a='b', field_b='c', field_c='a')

        class TestSortForm(SortForm):
            model = ChangeListModel
            HEADERS = (
                Header('field_a'),
                Header('field_b'),
                Header('named_header', column_name='field_c'),
                Header('not_sortable', is_sortable=False),
            )
        self.TestSortForm = TestSortForm

    def test_valid_with_no_sorts(self):
        form = self.TestSortForm({})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.get_queryset().count(), 3)

    def test_sort_field_cleaning(self):
        self.assertTrue(self.TestSortForm({'sorts': '1.2.3'}).is_valid())
        self.assertTrue(self.TestSortForm({'sorts': '2.3.1'}).is_valid())
        self.assertTrue(self.TestSortForm({'sorts': '3.1.2'}).is_valid())

        # Unsortable Header
        unsortable = self.TestSortForm({'sorts': '1.2.3.4'})
        self.assertFalse(unsortable.is_valid())
        self.assertIn('sorts', unsortable.errors)
        self.assertIn(self.TestSortForm.error_messages['unsortable_header'], unsortable.errors['sorts'])

        # Unknown Header
        unknown = self.TestSortForm({'sorts': '1.2.3.5'})
        self.assertFalse(unknown.is_valid())
        self.assertIn('sorts', unsortable.errors)
        self.assertIn(self.TestSortForm.error_messages['unknown_header'], unsortable.errors['sorts'])

        # Invalid Header
        unknown = self.TestSortForm({'sorts': '1.2.X'})
        self.assertFalse(unknown.is_valid())
        self.assertIn('sorts', unsortable.errors)
        self.assertIn(self.TestSortForm.error_messages['unknown_header'], unsortable.errors['sorts'])

    def test_single_field_sorting(self):
        f = self.TestSortForm({'sorts': '1'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.abc, self.bca, self.cab),
        )

        f = self.TestSortForm({'sorts': '-1'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.cab, self.bca, self.abc),
        )

        f = self.TestSortForm({'sorts': '2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.cab, self.abc, self.bca),
        )

        f = self.TestSortForm({'sorts': '-2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.bca, self.abc, self.cab),
        )

        f = self.TestSortForm({'sorts': '3'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.bca, self.cab, self.abc),
        )

        f = self.TestSortForm({'sorts': '-3'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.abc, self.cab, self.bca),
        )

    def test_multi_field_sorting(self):
        self.aac = ChangeListModel.objects.create(field_a='a', field_b='a', field_c='c')

        f = self.TestSortForm({'sorts': '1.2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.aac, self.abc, self.bca, self.cab),
        )

        f = self.TestSortForm({'sorts': '1.-2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.abc, self.aac, self.bca, self.cab),
        )

    def test_order_by_override(self):
        self.aac = ChangeListModel.objects.create(field_a='a', field_b='a', field_c='c')
        self.aab = ChangeListModel.objects.create(field_a='a', field_b='a', field_c='b')

        class OverriddenOrderForm(self.TestSortForm):
            def get_order_by(self):
                order_by = super(OverriddenOrderForm, self).get_order_by()
                return ['field_a'] + order_by

        f = OverriddenOrderForm({'sorts': '-3.2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.aac, self.abc, self.aab, self.bca, self.cab),
        )

        f = OverriddenOrderForm({'sorts': '3.2'})
        f.full_clean()
        self.assertSequenceEqual(
            f.get_queryset(),
            (self.aab, self.aac, self.abc, self.bca, self.cab),
        )



#!/usr/bin/env python
from setuptools import setup, find_packages
import os

__doc__ = """
App for Django featuring improved form base classes.
"""

version = '1.2.2'


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='yl-django-betterforms',
    version=version,
    description=__doc__,
    url="https://django-betterforms.readthedocs.org/en/latest/",
    author="Fusionbox",
    author_email='programmers@fusionbox.com',
    packages=[package for package in find_packages()
              if package.startswith('betterforms')],
    install_requires=['Django>=1.8', 'six'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



from itertools import chain
from operator import add

try:
    from collections import OrderedDict
except ImportError:  # Python 2.6, Django < 1.7
    from django.utils.datastructures import SortedDict as OrderedDict  # NOQA

try:
    from django.forms.utils import ErrorDict, ErrorList
except ImportError:  # Django < 1.7
    from django.forms.util import ErrorDict, ErrorList  # NOQA

try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError:  # Python < 3.0, Django < 3.0
    def python_2_unicode_compatible(klass):
        return klass

from django.core.exceptions import ValidationError, NON_FIELD_ERRORS
from django.utils.safestring import mark_safe
from six.moves import reduce


@python_2_unicode_compatible
class MultiForm(object):
    """
    A container that allows you to treat multiple forms as one form.  This is
    great for using more than one form on a page that share the same submit
    button.  MultiForm imitates the Form API so that it is invisible to anybody
    else that you are using a MultiForm.
    """
    form_classes = {}

    def __init__(self, data=None, files=None, *args, **kwargs):
        # Some things, such as the WizardView expect these to exist.
        self.data, self.files = data, files
        kwargs.update(
            data=data,
            files=files,
        )

        self.initials = kwargs.pop('initial', None)
        if self.initials is None:
            self.initials = {}
        self.forms = OrderedDict()
        self.crossform_errors = []

        for key, form_class in self.form_classes.items():
            fargs, fkwargs = self.get_form_args_kwargs(key, args, kwargs)
            self.forms[key] = form_class(*fargs, **fkwargs)

    def get_form_args_kwargs(self, key, args, kwargs):
        """
        Returns the args and kwargs for initializing one of our form children.
        """
        fkwargs = kwargs.copy()
        prefix = kwargs.get('prefix')
        if prefix is None:
            prefix = key
        else:
            prefix = '{0}__{1}'.format(key, prefix)
        fkwargs.update(
            initial=self.initials.get(key),
            prefix=prefix,
        )
        return args, fkwargs

    def __str__(self):
        return self.as_table()

    def __getitem__(self, key):
        return self.forms[key]

    @property
    def errors(self):
        errors = {}
        for form_name in self.forms:
            form = self.forms[form_name]
            for field_name in form.errors:
                errors[form.add_prefix(field_name)] = form.errors[field_name]
        if self.crossform_errors:
            errors[NON_FIELD_ERRORS] = self.crossform_errors
        return errors

    @property
    def fields(self):
        fields = []
        for form_name in self.forms:
            form = self.forms[form_name]
            for field_name in form.fields:
                fields += [form.add_prefix(field_name)]
        return fields

    def __iter__(self):
        # TODO: Should the order of the fields be controllable from here?
        return chain.from_iterable(self.forms.values())

    @property
    def is_bound(self):
        return any(form.is_bound for form in self.forms.values())

    def clean(self):
        """
        Raises any ValidationErrors required for cross form validation. Should
        return a dict of cleaned_data objects for any forms whose data should
        be overridden.
        """
        return self.cleaned_data

    def add_crossform_error(self, e):
        self.crossform_errors.append(e)

    def is_valid(self):
        forms_valid = all(form.is_valid() for form in self.forms.values())
        try:
            self.cleaned_data = self.clean()
        except ValidationError as e:
            self.add_crossform_error(e)
        return forms_valid and not self.crossform_errors

    def non_field_errors(self):
        form_errors = (
            form.non_field_errors() for form in self.forms.values()
            if hasattr(form, 'non_field_errors')
        )
        return ErrorList(chain(self.crossform_errors, *form_errors))

    def as_table(self):
        return mark_safe(''.join(form.as_table() for form in self.forms.values()))

    def as_ul(self):
        return mark_safe(''.join(form.as_ul() for form in self.forms.values()))

    def as_p(self):
        return mark_safe(''.join(form.as_p() for form in self.forms.values()))

    def is_multipart(self):
        return any(form.is_multipart() for form in self.forms.values())

    @property
    def media(self):
        return reduce(add, (form.media for form in self.forms.values()))

    def hidden_fields(self):
        # copy implementation instead of delegating in case we ever
        # want to override the field ordering.
        return [field for field in self if field.is_hidden]

    def visible_fields(self):
        return [field for field in self if not field.is_hidden]

    @property
    def cleaned_data(self):
        return OrderedDict(
            (key, form.cleaned_data)
            for key, form in self.forms.items() if form.is_valid()
        )

    @cleaned_data.setter
    def cleaned_data(self, data):
        for key, value in data.items():
            child_form = self[key]
            if hasattr(child_form, 'forms'):
                for formlet, formlet_data in zip(child_form.forms, value):
                    formlet.cleaned_data = formlet_data
            else:
                child_form.cleaned_data = value


class MultiModelForm(MultiForm):
    """
    MultiModelForm adds ModelForm support on top of MultiForm.  That simply
    means that it includes support for the instance parameter in initialization
    and adds a save method.
    """
    def __init__(self, *args, **kwargs):
        self.instances = kwargs.pop('instance', None)
        if self.instances is None:
            self.instances = {}
        super(MultiModelForm, self).__init__(*args, **kwargs)

    def get_form_args_kwargs(self, key, args, kwargs):
        fargs, fkwargs = super(MultiModelForm, self).get_form_args_kwargs(key, args, kwargs)
        try:
            # If we only pass instance when there was one specified, we make it
            # possible to use non-ModelForms together with ModelForms.
            fkwargs['instance'] = self.instances[key]
        except KeyError:
            pass
        return fargs, fkwargs

    def save(self, commit=True):
        objects = OrderedDict(
            (key, form.save(commit))
            for key, form in self.forms.items()
        )

        if any(hasattr(form, 'save_m2m') for form in self.forms.values()):
            def save_m2m():
                for form in self.forms.values():
                    if hasattr(form, 'save_m2m'):
                        form.save_m2m()
            self.save_m2m = save_m2m

        return objects



from django.views.generic import ListView, FormView


class BrowseView(ListView, FormView):
    """
    Class Based view for working with changelists.
    """
    def post(self, *args, **kwargs):
        return self.http_method_not_allowed(*args, **kwargs)

    def get_form_kwargs(self):
        kwargs = {
            'initial': self.get_initial(),
            'queryset': self.object_list,
            'data': self.request.GET,
            'files': self.request.FILES,
        }
        return kwargs

    def get_context_data(self, **kwargs):
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        kwargs['form'] = form
        if form.is_valid():
            kwargs['object_list'] = form.get_queryset()
        else:
            kwargs['object_list'] = form.base_queryset.none()
        kwargs = super(BrowseView, self).get_context_data(**kwargs)
        return kwargs
